

/*
 * ====================== WARNING ======================
 *
 * THE CONTENTS OF THIS FILE HAVE BEEN AUTO-GENERATED.
 * DO NOT MODIFY IN ANY WAY.
 *
 * ====================== WARNING ======================
 */

/**
 * @page c_le_avdata_legacy AirVantage Data
 *
 * @deprecated This API should not be used for new applications and will be removed in a future
 * version of Legato. It has been replaced by @ref c_le_avdata .
 *
 * @ref le_avdata_interface.h "API Reference" <br>
 * @ref howToAVData "How To Manage Data"
 *
 * <HR>
 *
 * This API provides a data service to allow apps to manage app-specific
 * data on the AirVantage server.
 *
 * Data is setup as @b assets &mdash;  a collection of fields that can be managed by the AirVantage
 * server.
 *
 * An asset field is a single piece of information that can be managed by the AirVantage server.
 * There can be multiple fields in an asset.
 *
 * A field can be:
 * - @c variable allows an app to set the value, and can be read from the AV server.
 * - @c setting allows the AV server to set the value, and can be read by an app.
 * - @c command allows the AV server to invoke a command or function in the app.
 *
 * All fields have names. Variable and setting fields also have a type. The possible field types,
 * along with the default value are:
 * - string (empty string)
 * - integer (0)
 * - float (0.0)
 * - boolean (false)
 * - binary (zero-length block).
 *
 * Default values can be overwritten in the asset definition. Currently, string, integer, float
 * and boolean fields are supported; binary fields will be supported in the future.
 *
 *
 * @section le_avdata_instance Asset Data Instances
 *
 * An app that needs to send data to the AirVantage server must first create an asset instance
 * using le_avc_Create() with the name of the asset definition.
 * Once an asset instance is created, the app can begin accessing the instance's fields.
 *
 * Multiple instances of the same asset can be created, as well as multiple instances of different
 * assets.
 *
 * Asset definitions are specified in the @ref defFilesCdef_assets section of the app's @c cdef file.
 *
 * @section le_avdata_field Field Values and Activity
 *
 * Set functions are available to set variable field values. Get functions are
 * available to get settings fields' values.
 *
 * An app can register a handler so that it can be called when activity occurs on a field.
 * This is optional for variable and setting fields, but is required for command fields.
 * - @c variable called when the field is read by the AV server. The
 * app can then call the appropriate set function to provide a new value for the field.
 * - @c setting called when the field has been updated by the AV server
 * with a new value. The app can use the appropriate get function to retrieve the new value.
 * - @c command called when the AV server wants to invoke a command. The
 * app should perform an appropriate action or call a function to execute this command.
 *
 * Read action handlers are only called when the server performs the read on a specific
 * field of an asset instance. i.e, a handler registered on a variable field will
 * be called if the server reads only that field; it will not be called when the server
 * reads the whole instance of an asset. Write action handlers are called on write
 * to asset instance or individual fields.
 *
 * An app can request to open avms session and register a handler to get notified of avms session
 * events. le_avdata_RequestSession() and le_avdata_ReleaseSession() can be used to
 * open an avms session and close an avms session respectively. If the session was initiated by an
 * user app, avms session will be closed when all apps release their session reference.
 * le_avdata_AddSessionStateHandler() and le_avdata_RemoveSessionStateHandler() can be used to add
 * and remove notification handlers.
 *
 * Leaving it optional to register handlers for variable and setting fields allows an
 * app to decide how it wants to access variable and setting fields. It can decide to
 * only do something in response to the AV server, or it can work independently of the AV server
 * updating variables when it has a new value, and reading settings only when it needs the value.
 *
 * @section le_avdata_observe Observe
 *
 * If the @ref howToAVData_enableObserve "Observe" is enabled, the AirVantage Server will send
 * observe requests on all user-defined assets of an app.
 *
 * Whenever an app (asset) changes a field value by using le_avdata_Set*(), it'll trigger a
 * notify if Observe is enabled on that asset. The notify contains only the value of the changed
 * field.
 *
 * @section le_avdata_timeseries Time Series
 *
 * Time series is an AirVantage-specific LWM2M feature built on top of LWM2M Observe.
 * This feature enables user apps to collect sensor data when the device is offline. You control
 * starting and stopping time series data on a resource through the user app.
 * le_avdata_StartTimeSeries() starts time series data on a resource, and le_avdata_StopTimeSeries()
 * stops collecting time series data on a resource. User apps can open an @c avms session, and push the
 * collected history data using le_avdata_PushTimeSeries().
 *
 * The buffer size allocated for history data per resource is 1024 bytes. Bytes transmitted
 * over the air can be reduced by choosing an appropriate factor. For example, if the sampled
 * integer data is a multiple of 1000, the encoded data will be smaller if a factor of 0.001 is
 * used. For float fields, if a factor other than 1 is used, the data will be encoded as integer to save
 * bytes transported over the air. For example, if the resolution of float data is 0.01, a factor of
 * 100 can be used to represent .01 as 1, and encoding this as integer thus saving memory.
 *
 * @note System time is used as timestamp for history data in le_avdata_Set*(). It's up to the
 * target device administrator to ensure system time is up-to-date before starting time series.
 * Alternatively, le_avdata_Record*() can be used to pass a user-specified timestamp.
 *
 * le_avdata_RecordInt(), le_avdata_RecordFloat(), le_avdata_RecordBool() and
 * le_avdata_RecordString() can be used to pass an user specified time stamp. The user specified
 * time stamp must be in milli seconds elapsed since epoch.
 *
 * @note Observe has to be enabled on the resource before time series can be pushed out. User apps can
 * use le_avdata_IsObserve() to know if Observe is enabled on a resource.
 *
 * @note LEGATO_FEATURE_TIMESERIES is enabled by default for all WP products. The time series
 * feature depends on the cbor libraries in yocto. Hence, yocto image has to
 * be updated on the device together with the legato image. Users who want to update
 * legato only without the Yocto image should turn off time series in targetdefs.
 *
 * @section le_avdata_fatal Fatal Behavior
 *
 * An invalid asset name or field name is treated as a fatal error (i.e. non-recoverable)
 * and will result in the client app being terminated.
 *
 * <HR>
 *
 * Copyright (C) Sierra Wireless Inc.
 */
/**
 * @file le_avdata_interface.h
 *
 * Legato @ref c_le_avdata include file.
 *
 * Copyright (C) Sierra Wireless Inc.
 */

#ifndef LE_AVDATA_INTERFACE_H_INCLUDE_GUARD
#define LE_AVDATA_INTERFACE_H_INCLUDE_GUARD


#include "legato.h"


//--------------------------------------------------------------------------------------------------
/**
 * Type for handler called when a server disconnects.
 */
//--------------------------------------------------------------------------------------------------
typedef void (*le_avdata_DisconnectHandler_t)(void *);

//--------------------------------------------------------------------------------------------------
/**
 *
 * Connect the current client thread to the service providing this API. Block until the service is
 * available.
 *
 * For each thread that wants to use this API, either ConnectService or TryConnectService must be
 * called before any other functions in this API.  Normally, ConnectService is automatically called
 * for the main thread, but not for any other thread. For details, see @ref apiFilesC_client.
 *
 * This function is created automatically.
 */
//--------------------------------------------------------------------------------------------------
void le_avdata_ConnectService
(
    void
);

//--------------------------------------------------------------------------------------------------
/**
 *
 * Try to connect the current client thread to the service providing this API. Return with an error
 * if the service is not available.
 *
 * For each thread that wants to use this API, either ConnectService or TryConnectService must be
 * called before any other functions in this API.  Normally, ConnectService is automatically called
 * for the main thread, but not for any other thread. For details, see @ref apiFilesC_client.
 *
 * This function is created automatically.
 *
 * @return
 *  - LE_OK if the client connected successfully to the service.
 *  - LE_UNAVAILABLE if the server is not currently offering the service to which the client is
 *    bound.
 *  - LE_NOT_PERMITTED if the client interface is not bound to any service (doesn't have a binding).
 *  - LE_COMM_ERROR if the Service Directory cannot be reached.
 */
//--------------------------------------------------------------------------------------------------
le_result_t le_avdata_TryConnectService
(
    void
);

//--------------------------------------------------------------------------------------------------
/**
 * Set handler called when server disconnection is detected.
 *
 * When a server connection is lost, call this handler then exit with LE_FATAL.  If a program wants
 * to continue without exiting, it should call longjmp() from inside the handler.
 */
//--------------------------------------------------------------------------------------------------
void le_avdata_SetServerDisconnectHandler
(
    le_avdata_DisconnectHandler_t disconnectHandler,
    void *contextPtr
);

//--------------------------------------------------------------------------------------------------
/**
 *
 * Disconnect the current client thread from the service providing this API.
 *
 * Normally, this function doesn't need to be called. After this function is called, there's no
 * longer a connection to the service, and the functions in this API can't be used. For details, see
 * @ref apiFilesC_client.
 *
 * This function is created automatically.
 */
//--------------------------------------------------------------------------------------------------
void le_avdata_DisconnectService
(
    void
);


//--------------------------------------------------------------------------------------------------
/**
 * Define the maximum length of an asset name (TODO: verify value)
 */
//--------------------------------------------------------------------------------------------------
#define LE_AVDATA_ASSET_NAME_LEN 47

//--------------------------------------------------------------------------------------------------
/**
 * Define the maximum length of a field name (TODO: verify value)
 */
//--------------------------------------------------------------------------------------------------
#define LE_AVDATA_FIELD_NAME_LEN 47

//--------------------------------------------------------------------------------------------------
/**
 * Define the maximum length of a string field (TODO: verify value)
 */
//--------------------------------------------------------------------------------------------------
#define LE_AVDATA_STRING_VALUE_LEN 255

//--------------------------------------------------------------------------------------------------
/**
 * Define the maximum length of a binary data field (TODO: verify value)
 */
//--------------------------------------------------------------------------------------------------
#define LE_AVDATA_BINARY_VALUE_LEN 255

//--------------------------------------------------------------------------------------------------
/**
 * An asset data instance reference
 */
//--------------------------------------------------------------------------------------------------
typedef struct le_avdata_AssetInstance* le_avdata_AssetInstanceRef_t;


//--------------------------------------------------------------------------------------------------
/**
 * Reference returned by RequestSession function and used by ReleaseSession function
 */
//--------------------------------------------------------------------------------------------------
typedef struct le_avdata_RequestSessionObj* le_avdata_RequestSessionObjRef_t;


//--------------------------------------------------------------------------------------------------
/**
 * AVMS session state
 */
//--------------------------------------------------------------------------------------------------
typedef enum
{
    LE_AVDATA_SESSION_STARTED = 0,
        ///< AVMS session started
    LE_AVDATA_SESSION_STOPPED = 1
        ///< AVMS session stopped
}
le_avdata_SessionState_t;


//--------------------------------------------------------------------------------------------------
/**
 * Reference type used by Add/Remove functions for EVENT 'le_avdata_FieldEvent'
 */
//--------------------------------------------------------------------------------------------------
typedef struct le_avdata_FieldEventHandler* le_avdata_FieldEventHandlerRef_t;


//--------------------------------------------------------------------------------------------------
/**
 * Reference type used by Add/Remove functions for EVENT 'le_avdata_SessionState'
 */
//--------------------------------------------------------------------------------------------------
typedef struct le_avdata_SessionStateHandler* le_avdata_SessionStateHandlerRef_t;


//--------------------------------------------------------------------------------------------------
/**
 * Handler for field activity
 */
//--------------------------------------------------------------------------------------------------
typedef void (*le_avdata_FieldHandlerFunc_t)
(
    le_avdata_AssetInstanceRef_t instRef,
        ///<
    const char* LE_NONNULL fieldName,
        ///<
    void* contextPtr
        ///<
);

//--------------------------------------------------------------------------------------------------
/**
 * Handler for AV session changes
 */
//--------------------------------------------------------------------------------------------------
typedef void (*le_avdata_SessionStateHandlerFunc_t)
(
    le_avdata_SessionState_t sessionState,
        ///< Session started or stopped?
    void* contextPtr
        ///<
);

//--------------------------------------------------------------------------------------------------
/**
 * Add handler function for EVENT 'le_avdata_FieldEvent'
 *
 * This event provides information on field activity
 */
//--------------------------------------------------------------------------------------------------
le_avdata_FieldEventHandlerRef_t le_avdata_AddFieldEventHandler
(
    le_avdata_AssetInstanceRef_t instRef,
        ///< [IN]
    const char* LE_NONNULL fieldName,
        ///< [IN]
    le_avdata_FieldHandlerFunc_t handlerPtr,
        ///< [IN]
    void* contextPtr
        ///< [IN]
);

//--------------------------------------------------------------------------------------------------
/**
 * Remove handler function for EVENT 'le_avdata_FieldEvent'
 */
//--------------------------------------------------------------------------------------------------
void le_avdata_RemoveFieldEventHandler
(
    le_avdata_FieldEventHandlerRef_t handlerRef
        ///< [IN]
);

//--------------------------------------------------------------------------------------------------
/**
 * Create an instance of AirVantage asset
 *
 * @return Reference to the asset instance
 *
 * @note It is a fatal error if the asset isn't defined.
 */
//--------------------------------------------------------------------------------------------------
le_avdata_AssetInstanceRef_t le_avdata_Create
(
    const char* LE_NONNULL assetName
        ///< [IN]
);

//--------------------------------------------------------------------------------------------------
/**
 * Delete an instance of AirVantage asset
 *
 * @note It's a fatal error if the instRef isn't valid
 */
//--------------------------------------------------------------------------------------------------
void le_avdata_Delete
(
    le_avdata_AssetInstanceRef_t instRef
        ///< [IN]
);

//--------------------------------------------------------------------------------------------------
/**
 * Get the value of an integer setting field
 *
 * @note The client will be terminated if the instRef is not valid, or the field doesn't exist
 */
//--------------------------------------------------------------------------------------------------
void le_avdata_GetInt
(
    le_avdata_AssetInstanceRef_t instRef,
        ///< [IN]
    const char* LE_NONNULL fieldName,
        ///< [IN]
    int32_t* valuePtr
        ///< [OUT]
);

//--------------------------------------------------------------------------------------------------
/**
 * Set the value of an integer variable field
 *
 * @note The client will be terminated if the instRef is not valid, or the field doesn't exist
 *
 * @return:
 *      - LE_OK on success
 *      - LE_OVERFLOW if the current entry was NOT added as the time series buffer is full.
 *                    (This error is applicable only if time series is enabled on this field)
 *      - LE_NO_MEMORY if the current entry was added but there is no space for next one.
 *                    (This error is applicable only if time series is enabled on this field)
 *      - LE_FAULT on any other error
 */
//--------------------------------------------------------------------------------------------------
le_result_t le_avdata_SetInt
(
    le_avdata_AssetInstanceRef_t instRef,
        ///< [IN]
    const char* LE_NONNULL fieldName,
        ///< [IN]
    int32_t value
        ///< [IN]
);

//--------------------------------------------------------------------------------------------------
/**
 * Get the value of a float setting field
 *
 * @note The client will be terminated if the instRef is not valid, or the field doesn't exist
 */
//--------------------------------------------------------------------------------------------------
void le_avdata_GetFloat
(
    le_avdata_AssetInstanceRef_t instRef,
        ///< [IN]
    const char* LE_NONNULL fieldName,
        ///< [IN]
    double* valuePtr
        ///< [OUT]
);

//--------------------------------------------------------------------------------------------------
/**
 * Set the value of a float variable field
 *
 * @note The client will be terminated if the instRef is not valid, or the field doesn't exist
 *
 * @return:
 *      - LE_OK on success
 *      - LE_OVERFLOW if the current entry was NOT added as the time series buffer is full.
 *                    (This error is applicable only if time series is enabled on this field)
 *      - LE_NO_MEMORY if the current entry was added but there is no space for next one.
 *                    (This error is applicable only if time series is enabled on this field)
 *      - LE_FAULT on any other error
 */
//--------------------------------------------------------------------------------------------------
le_result_t le_avdata_SetFloat
(
    le_avdata_AssetInstanceRef_t instRef,
        ///< [IN]
    const char* LE_NONNULL fieldName,
        ///< [IN]
    double value
        ///< [IN]
);

//--------------------------------------------------------------------------------------------------
/**
 * Get the value of a boolean setting field
 *
 * @note The client will be terminated if the instRef is not valid, or the field doesn't exist
 */
//--------------------------------------------------------------------------------------------------
void le_avdata_GetBool
(
    le_avdata_AssetInstanceRef_t instRef,
        ///< [IN]
    const char* LE_NONNULL fieldName,
        ///< [IN]
    bool* valuePtr
        ///< [OUT]
);

//--------------------------------------------------------------------------------------------------
/**
 * Set the value of a boolean variable field
 *
 * @note The client will be terminated if the instRef is not valid, or the field doesn't exist
 *
 * @return:
 *      - LE_OK on success
 *      - LE_OVERFLOW if the current entry was NOT added as the time series buffer is full.
 *                    (This error is applicable only if time series is enabled on this field)
 *      - LE_NO_MEMORY if the current entry was added but there is no space for next one.
 *                    (This error is applicable only if time series is enabled on this field)
 *      - LE_FAULT on any other error
 */
//--------------------------------------------------------------------------------------------------
le_result_t le_avdata_SetBool
(
    le_avdata_AssetInstanceRef_t instRef,
        ///< [IN]
    const char* LE_NONNULL fieldName,
        ///< [IN]
    bool value
        ///< [IN]
);

//--------------------------------------------------------------------------------------------------
/**
 * Get the value of a string setting field
 *
 * @note The client will be terminated if the instRef is not valid, or the field doesn't exist
 */
//--------------------------------------------------------------------------------------------------
void le_avdata_GetString
(
    le_avdata_AssetInstanceRef_t instRef,
        ///< [IN]
    const char* LE_NONNULL fieldName,
        ///< [IN]
    char* value,
        ///< [OUT]
    size_t valueSize
        ///< [IN]
);

//--------------------------------------------------------------------------------------------------
/**
 * Set the value of a string variable field
 *
 * @note The client will be terminated if the instRef is not valid, or the field doesn't exist
 *
 * @return:
 *      - LE_OK on success
 *      - LE_OVERFLOW if the stored string was truncated or
 *                    if the current entry was NOT added as the time series buffer is full.
 *      - LE_NO_MEMORY if the current entry was added but there is no space for next one.
 *                    (This error is applicable only if time series is enabled on this field)
 *      - LE_FAULT on any other error
 */
//--------------------------------------------------------------------------------------------------
le_result_t le_avdata_SetString
(
    le_avdata_AssetInstanceRef_t instRef,
        ///< [IN]
    const char* LE_NONNULL fieldName,
        ///< [IN]
    const char* LE_NONNULL value
        ///< [IN]
);

//--------------------------------------------------------------------------------------------------
/**
 * Get the value of a binary data setting field
 *
 * @note The client will be terminated if the instRef is not valid, or the field doesn't exist
 */
//--------------------------------------------------------------------------------------------------
void le_avdata_GetBinary
(
    le_avdata_AssetInstanceRef_t instRef,
        ///< [IN]
    const char* LE_NONNULL fieldName,
        ///< [IN]
    uint8_t* valuePtr,
        ///< [OUT]
    size_t* valueSizePtr
        ///< [INOUT]
);

//--------------------------------------------------------------------------------------------------
/**
 * Set the value of a binary data variable field
 *
 * @note The client will be terminated if the instRef is not valid, or the field doesn't exist
 */
//--------------------------------------------------------------------------------------------------
void le_avdata_SetBinary
(
    le_avdata_AssetInstanceRef_t instRef,
        ///< [IN]
    const char* LE_NONNULL fieldName,
        ///< [IN]
    const uint8_t* valuePtr,
        ///< [IN]
    size_t valueSize
        ///< [IN]
);

//--------------------------------------------------------------------------------------------------
/**
 * Allocate resources and set up cbor stream so that we can start accumulating time series
 * data on the specified field.
 *
 * @note client will be terminated if instRef isn't valid, or the field doesn't exist
 *
 * @note Currently the buffer size of time series data is limited to 1024 bytes. When the buffer
 *       overflows the device has to push the buffer before recording new entries.
 *
 * @note Factor is applicable only for integer and float fields. For all other fields factor will be
 *       silently ignored. Also a factor of "0" will be ignored for integer resources. The factor
 *       will be applied (multiplication) before encoding the sample.
 *
 * @return
 *      - LE_OK on success
 *      - LE_BUSY if time series already enabled on this field
 *      - LE_FAULT on any other error
 */
//--------------------------------------------------------------------------------------------------
le_result_t le_avdata_StartTimeSeries
(
    le_avdata_AssetInstanceRef_t instRef,
        ///< [IN]
    const char* LE_NONNULL fieldName,
        ///< [IN]
    double factor,
        ///< [IN]
    double timeStampFactor
        ///< [IN]
);

//--------------------------------------------------------------------------------------------------
/**
 * Stop time series on this field and free resources.
 *
 * @note client will be terminated if instRef isn't valid, or the field doesn't exist
 *
 * @return
 *      - LE_OK on success
 *      - LE_CLOSED if timeseries already stopped
 */
//--------------------------------------------------------------------------------------------------
le_result_t le_avdata_StopTimeSeries
(
    le_avdata_AssetInstanceRef_t instRef,
        ///< [IN]
    const char* LE_NONNULL fieldName
        ///< [IN]
);

//--------------------------------------------------------------------------------------------------
/**
 * Compress the accumulated CBOR encoded time series data and send it to server. After the data is
 * pushed, time series will be stopped and not started again unless isRestartTimeSeries is true.
 *
 * @note client will be terminated if instRef isn't valid, or the field doesn't exist
 *
 * @return:
 *      - LE_OK on success
 *      - LE_CLOSED if time series not enabled on this field
 *      - LE_UNAVAILABLE if observe is not enabled on this field
 *      - LE_FAULT if any other error
 */
//--------------------------------------------------------------------------------------------------
le_result_t le_avdata_PushTimeSeries
(
    le_avdata_AssetInstanceRef_t instRef,
        ///< [IN]
    const char* LE_NONNULL fieldName,
        ///< [IN]
    bool isRestartTimeSeries
        ///< [IN]
);

//--------------------------------------------------------------------------------------------------
/**
 * Add handler function for EVENT 'le_avdata_SessionState'
 *
 * This event provides information on AV session state changes
 */
//--------------------------------------------------------------------------------------------------
le_avdata_SessionStateHandlerRef_t le_avdata_AddSessionStateHandler
(
    le_avdata_SessionStateHandlerFunc_t handlerPtr,
        ///< [IN]
    void* contextPtr
        ///< [IN]
);

//--------------------------------------------------------------------------------------------------
/**
 * Remove handler function for EVENT 'le_avdata_SessionState'
 */
//--------------------------------------------------------------------------------------------------
void le_avdata_RemoveSessionStateHandler
(
    le_avdata_SessionStateHandlerRef_t handlerRef
        ///< [IN]
);

//--------------------------------------------------------------------------------------------------
/**
 * Is time series enabled on this resource, if yes how many data points are recorded so far?
 *
 * @return
 *      - LE_OK on success
 *      - LE_FAULT if time series not supported
 */
//--------------------------------------------------------------------------------------------------
le_result_t le_avdata_GetTimeSeriesStatus
(
    le_avdata_AssetInstanceRef_t instRef,
        ///< [IN]
    const char* LE_NONNULL fieldName,
        ///< [IN]
    bool* isTimeSeriesPtr,
        ///< [OUT]
    int32_t* numDataPointPtr
        ///< [OUT]
);

//--------------------------------------------------------------------------------------------------
/**
 * Record the value of an integer variable field in time series.
 *
 * @note The client will be terminated if the instRef is not valid, or the field doesn't exist
 *
 * @note This function is the same as the SetInt() except that it provides an option to pass the
 *       timestamp. SetInt() can be used to record time series with system time as the timestamp.
 *       Timestamp should be in milli seconds elapsed since epoch.
 *
 * @return:
 *      - LE_OK on success
 *      - LE_OVERFLOW if the current entry was NOT added as the time series buffer is full.
 *                    (This error is applicable only if time series is enabled on this field)
 *      - LE_NO_MEMORY if the current entry was added but there is no space for next one.
 *                    (This error is applicable only if time series is enabled on this field)
 *      - LE_FAULT on any other error
 */
//--------------------------------------------------------------------------------------------------
le_result_t le_avdata_RecordInt
(
    le_avdata_AssetInstanceRef_t instRef,
        ///< [IN]
    const char* LE_NONNULL fieldName,
        ///< [IN]
    int32_t value,
        ///< [IN]
    uint64_t timestamp
        ///< [IN]
);

//--------------------------------------------------------------------------------------------------
/**
 * Record the value of a float variable field in time series.
 *
 * @note The client will be terminated if the instRef is not valid, or the field doesn't exist
 *
 * @note This function is the same as the SetFloat() except that it provides an option to pass the
 *       timestamp. SetFloat() can be used to record time series with system time as the timestamp.
 *       Timestamp should be in milli seconds elapsed since epoch.
 *
 * @return:
 *      - LE_OK on success
 *      - LE_OVERFLOW if the current entry was NOT added as the time series buffer is full.
 *                    (This error is applicable only if time series is enabled on this field)
 *      - LE_NO_MEMORY if the current entry was added but there is no space for next one.
 *                    (This error is applicable only if time series is enabled on this field)
 *      - LE_FAULT on any other error
 */
//--------------------------------------------------------------------------------------------------
le_result_t le_avdata_RecordFloat
(
    le_avdata_AssetInstanceRef_t instRef,
        ///< [IN]
    const char* LE_NONNULL fieldName,
        ///< [IN]
    double value,
        ///< [IN]
    uint64_t timestamp
        ///< [IN]
);

//--------------------------------------------------------------------------------------------------
/**
 * Record the value of a boolean variable field in time series.
 *
 * @note The client will be terminated if the instRef is not valid, or the field doesn't exist
 *
 * @note This function is the same as the SetBool() except that it provides an option to pass the
 *       timestamp. SetBool() can be used to record time series with system time as the timestamp.
 *       Timestamp should be in milli seconds elapsed since epoch.
 *
 * @return:
 *      - LE_OK on success
 *      - LE_OVERFLOW if the current entry was NOT added as the time series buffer is full.
 *                    (This error is applicable only if time series is enabled on this field)
 *      - LE_NO_MEMORY if the current entry was added but there is no space for next one.
 *                    (This error is applicable only if time series is enabled on this field)
 *      - LE_FAULT on any other error
 */
//--------------------------------------------------------------------------------------------------
le_result_t le_avdata_RecordBool
(
    le_avdata_AssetInstanceRef_t instRef,
        ///< [IN]
    const char* LE_NONNULL fieldName,
        ///< [IN]
    bool value,
        ///< [IN]
    uint64_t timeStamp
        ///< [IN]
);

//--------------------------------------------------------------------------------------------------
/**
 * Record the value of a string variable field in time series
 *
 * @note The client will be terminated if the instRef is not valid, or the field doesn't exist
 *
 * @note This function is the same as the SetString() except that it provides an option to pass the
 *       timestamp. SetString() can be used to record time series with system time as the timestamp.
 *       Timestamp should be in milli seconds elapsed since epoch.
 *
 * @return:
 *      - LE_OK on success
 *      - LE_OVERFLOW if the stored string was truncated or
 *                    if the current entry was NOT added as the time series buffer is full.
 *      - LE_NO_MEMORY if the current entry was added but there is no space for next one.
 *                    (This error is applicable only if time series is enabled on this field)
 *      - LE_FAULT on any other error
 */
//--------------------------------------------------------------------------------------------------
le_result_t le_avdata_RecordString
(
    le_avdata_AssetInstanceRef_t instRef,
        ///< [IN]
    const char* LE_NONNULL fieldName,
        ///< [IN]
    const char* LE_NONNULL value,
        ///< [IN]
    uint64_t timeStamp
        ///< [IN]
);

//--------------------------------------------------------------------------------------------------
/**
 * Is this resource enabled for observe notifications?
 *
 * @note client will be terminated if instRef isn't valid, or the field doesn't exist
 *
 * @return
 *      - LE_OK on success
 *      - LE_NOT_FOUND if field not found
 */
//--------------------------------------------------------------------------------------------------
le_result_t le_avdata_IsObserve
(
    le_avdata_AssetInstanceRef_t instRef,
        ///< [IN]
    const char* LE_NONNULL fieldName,
        ///< [IN]
    bool* isObservePtr
        ///< [OUT]
);

//--------------------------------------------------------------------------------------------------
/**
 * Request the avcServer to open a session.
 *
 * @return
 *      - SessionRef if session request succeeded
 *      - NULL if session request failed
 */
//--------------------------------------------------------------------------------------------------
le_avdata_RequestSessionObjRef_t le_avdata_RequestSession
(
    void
);

//--------------------------------------------------------------------------------------------------
/**
 * Request the avcServer to close a session.
 *
 */
//--------------------------------------------------------------------------------------------------
void le_avdata_ReleaseSession
(
    le_avdata_RequestSessionObjRef_t requestRef
        ///< [IN] Reference to a previously opened AV session.
);

#endif // LE_AVDATA_INTERFACE_H_INCLUDE_GUARD